from .round_wrapper import *
from .preprocess import *
from .remove_redundancy import *
from .feature_selection import *
from .median_impute import *
from .visualizations import *
import os
import argparse
import time
import matplotlib.pyplot as plt
import warnings

plt.rcParams['pdf.fonttype'] = 42
plt.rcParams['font.family'] = 'Arial'


def read_psi_file(psi_file_path, n_metadata_cols):
    # read the psi file in and separate out the metadata columns from numeric PSI data
    psi, metadata = format_psi_matrix(psi_file_path, n=n_metadata_cols)
    metadata.index = psi.index
    psi = psi.dropna(how='all')
    metadata = metadata.loc[psi.index, :]

    return psi, metadata


def oncosplice_feature_selection(psi, metadata, fold_threshold, samples_differing, corr_threshold_intercorr, corr_n_events, corr_threshold, speed, write_files=True, savedir=None):
    # variance based feature selection
    metadata = variance_based_feature_selection(psi, metadata, fold_threshold=fold_threshold, samples_differing=samples_differing)

    metadata = metadata.loc[metadata['high_variance'] == True, :]
    psi = psi.loc[metadata.index, :]

    # inter-feature correlation module

    if speed == "yes":
        psi = fast_intercorrelation_based_feature_selection(psi, corr_threshold=corr_threshold_intercorr, corr_n_events=corr_n_events)

    else:
        print(
            "beginning intercorrelated based feature selection. This may take a while if input PSI file has more than 30,000 events (typical time 1 hour)...")
        psi = intercorrelation_based_feature_selection(psi, corr_threshold=corr_threshold_intercorr, corr_n_events=corr_n_events)

    psi = intercorrelation_based_feature_selection(psi, corr_threshold=corr_threshold_intercorr, corr_n_events=corr_n_events)

    metadata = metadata.loc[psi.index, :]

    # Find the non-redundant splicing events
    list_of_events = remove_redundant_splicing_events(psi, corr_threshold, metadata)
    list_of_events = list(set(list_of_events))

    # write the essential files out for future assessments
    variable_features = pd.DataFrame(list_of_events, columns=["EventID"])

    if write_files:
        variable_features.to_csv(os.path.join(savedir, "list_of_non_redundant_events.txt"), sep="\t")
        metadata.to_csv(os.path.join(savedir, "final_metadata.txt"), sep="\t")

    return variable_features, psi, metadata


def oncosplice(psi, metadata, variable_features, pca_corr_threshold, npcs, rank, force_broad, min_group_size, dPSI, dPSI_p_val, min_differential_events, top_n_differential_events, conservation, depletion_corr_threshold, speed, n_rounds, write_files, savedir):

    if rank == "k30":
        rank = 30
    else:
        rank = None
    print("NMF Rank used:")
    print(rank)

    if force_broad == "on":
        rank1 = 2
    else:
        rank1 = rank
    print("Force broad used:")
    print(rank1)

    print("number of rounds to iterate through:")
    print(n_rounds)

    # impute the psi matrix for certain downstream steps
    # formatted_psi_file_imp, metadata_imp = format_psi_matrix(imputed_psi_file_path)
    psi_imp = median_impute(psi)
    psi_imp.index = psi.index

    list_of_events = variable_features['EventID'].to_list()

    # Subset the PSI matrix to only the non-redundant events
    psi_0 = psi.loc[list_of_events, :]
    psi_imp_0 = psi_imp.loc[list_of_events, :]

    # Initialize an empty dictionaries to store DEG dataframes,
    final_clusters_dict = {}
    depleted_psi_file_after_round_imp_dict = {}
    depleted_psi_file_after_round_dict = {}
    deg_results_all_dict = {}

    print("Number of PCs used for feature selection")
    print(npcs)

    if write_files:
        os.chdir(savedir)

    # feature selection prior to Round 1

    pca_events_round = get_events_from_pca(psi_imp, psi, list_of_events, corr_threshold=pca_corr_threshold, n_components=npcs)

    print("STARING ROUND 1...")
    print("Number of events prior to entering ROUND 1: ")
    print(len(psi_0.index))

    # Round 1 OncoSplice
    final_clusters_i, depleted_psi_file_after_round_imp_i, depleted_psi_file_after_round_i, deg_results_all_i = round_wrapper(
        filename="Round1", full_psi_file=psi_0, full_imputed_psi_file=psi_imp_0, metadata=metadata,
        highly_variable_events=pca_events_round, rank=rank1, min_group_size=min_group_size, dPSI=dPSI, dPSI_p_val=dPSI_p_val,
        min_differential_events=min_differential_events, top_n_differential_events=top_n_differential_events, conservation=conservation, strictness="tough",
        depletion_corr_threshold=depletion_corr_threshold, write_files=write_files, speed_corr=speed)

    # Round i OncoSplice (tough)
    depleted_events_round_i = depleted_psi_file_after_round_i.index.to_list()

    print("Number of events removed after ROUND 1: ")
    print(len(psi_0.index) - len(depleted_psi_file_after_round_i.index))

    deg_results_all_i['cluster'] = 'R1_' + deg_results_all_i['cluster']
    # Add a meaninful prefix to all column names
    prefix = "R1_C"
    final_clusters_i = final_clusters_i.rename(columns=lambda x: prefix + str(x))

    # store the mandatory round 1 outputs in respective dictionaries
    key = "Round1"
    final_clusters_dict[key] = final_clusters_i
    # depleted_psi_file_after_round_imp_dict[key] = depleted_psi_file_after_round_imp_i
    # depleted_psi_file_after_round_dict[key] = depleted_psi_file_after_round_i
    deg_results_all_dict[key] = deg_results_all_i

    for round_j in range(2, n_rounds+1):

        # feature selection prior to Round J
        pca_events_round_j = get_events_from_pca(psi_imp, psi, depleted_events_round_i,
                                                 corr_threshold=pca_corr_threshold, n_components=npcs)

        print(f"STARING ROUND {round_j}...")
        print(f"Number of events prior to entering ROUND {round_j}: ")
        print(len(depleted_psi_file_after_round_i.index))

        # Round 2 Oncosplice
        final_clusters_j, depleted_psi_file_after_round_imp_j, depleted_psi_file_after_round_j, deg_results_all_j = round_wrapper(
            filename=f"Round{round_j}", full_psi_file=depleted_psi_file_after_round_i,
            full_imputed_psi_file=depleted_psi_file_after_round_imp_i, highly_variable_events=pca_events_round_j, metadata=metadata, rank=rank,
            min_group_size=min_group_size, dPSI=dPSI, dPSI_p_val=dPSI_p_val, min_differential_events=min_differential_events,
            top_n_differential_events=top_n_differential_events, conservation=conservation, strictness="tough",
            depletion_corr_threshold=depletion_corr_threshold, write_files=write_files, speed_corr=speed)

        # Round 3 OncoSplice
        depleted_events_round_j = depleted_psi_file_after_round_j.index.to_list()
        print(f"Number of events removed after ROUND {round_j}: ")
        print(len(depleted_psi_file_after_round_i.index) - len(depleted_psi_file_after_round_j.index))

        deg_results_all_j['cluster'] = f'R{round_j}_' + deg_results_all_j['cluster']
        # Add a meaninful prefix to all column names
        prefix = f"R{round_j}_C"
        final_clusters_j = final_clusters_j.rename(columns=lambda x: prefix + str(x))

        # store the outputs in respective dictionaries
        key = f"Round{round_j}"
        final_clusters_dict[key] = final_clusters_j
        # depleted_psi_file_after_round_imp_dict[key] = depleted_psi_file_after_round_imp_j
        # depleted_psi_file_after_round_dict[key] = depleted_psi_file_after_round_j
        deg_results_all_dict[key] = deg_results_all_j

        # reset vars for next round
        depleted_events_round_i = depleted_events_round_j
        depleted_psi_file_after_round_i = depleted_psi_file_after_round_j
        depleted_psi_file_after_round_imp_i = depleted_psi_file_after_round_imp_j

        if len(depleted_psi_file_after_round_i) < top_n_differential_events:
            warnings.warn(f"Less than 50 splicing events pending after round {round_j}. Halting further iterative clustering process...")
            break


    # final outputs
    final_clusters_all_rounds = pd.concat(final_clusters_dict.values(), axis=1)

    de_results_all_rounds = pd.concat(deg_results_all_dict.values(), axis=0, ignore_index=True)

    # Grouping by cluster assignment and type of feature, and counting occurrences
    grouped_de_df_event_annotation = de_results_all_rounds.groupby(['cluster', 'EventAnnotation']).size().unstack(fill_value=0)

    # Calculating percentage of each type of feature in each cluster
    grouped_event_annotation_percentage = grouped_de_df_event_annotation.div(grouped_de_df_event_annotation.sum(axis=1), axis=0) * 100

    # Grouping by cluster assignment and type of feature, and counting occurrences
    grouped_de_df_event_direction = de_results_all_rounds.groupby(['cluster', 'event_direction'])['ClusterID'].nunique().unstack(
        fill_value=0)

    # Calculating percentage of each type of feature in each cluster
    grouped_direction_annotation_percentage = grouped_de_df_event_direction.div(grouped_de_df_event_direction.sum(axis=1), axis=0) * 100


    if write_files:
        path = os.path.join(savedir, "FinalResults")
        os.mkdir(path)
        os.chdir(path)
        # Define color palette for the types of features
        colors = sns.color_palette("Set1")

        final_clusters_all_rounds.to_csv("MergedResults.txt", sep="\t")
        grouped_de_df_event_annotation.to_csv("event_annotations_numbers.txt", sep='\t')
        grouped_event_annotation_percentage.to_csv("event_annotations_percentage.txt", sep='\t')
        grouped_de_df_event_direction.to_csv("event_direction_numbers.txt", sep='\t')
        grouped_direction_annotation_percentage.to_csv("event_direction_percentage.txt", sep='\t')

        # Plotting horizontal stacked bar graph
        plt.figure(figsize=(10, 6))
        grouped_event_annotation_percentage.plot(kind='barh', stacked=True, color=colors)
        plt.title('Percentage of Features in Each Type by Cluster Assignment')
        plt.ylabel('Cluster Assignment')
        plt.xlabel('Percentage')
        plt.legend(title='Event Annotation')
        plt.tight_layout()  # Adjust layout to fit cleanly in PDF
        file_path = os.path.join(path, "event_annotations_percentage.pdf")
        plt.savefig(file_path, format='pdf', bbox_inches='tight')

        plt.figure(figsize=(10, 6))
        grouped_de_df_event_annotation.plot(kind='barh', stacked=True, color=colors)
        plt.title('Number of Features in Each Type by Cluster Assignment')
        plt.ylabel('Cluster Assignment')
        plt.xlabel('Number of Features')
        plt.legend(title='Event Annotation')
        plt.tight_layout()  # Adjust layout to fit cleanly in PDF
        file_path = os.path.join(path, "event_annotations_numbers.pdf")
        plt.savefig(file_path, format='pdf', bbox_inches='tight')

        # Plotting horizontal stacked bar graph
        plt.figure(figsize=(10, 6))
        grouped_direction_annotation_percentage.plot(kind='barh', stacked=True, color=colors)
        plt.title('Percentage of Cluster IDs in Each Type by Cluster/Group Assignment')
        plt.ylabel('Cluster Assignment')
        plt.xlabel('Percentage')
        plt.legend(title='Event Annotation')
        plt.tight_layout()  # Adjust layout to fit cleanly in PDF
        file_path = os.path.join(path, "event_direction_percentage.pdf")
        plt.savefig(file_path, format='pdf', bbox_inches='tight')

        plt.figure(figsize=(10, 6))
        grouped_de_df_event_direction.plot(kind='barh', stacked=True, color=colors)
        plt.title('Number of ClusterID in Each Type by Cluster Assignment')
        plt.ylabel('Cluster Assignment')
        plt.xlabel('Number of Features')
        plt.legend(title='Event Annotation')
        plt.tight_layout()  # Adjust layout to fit cleanly in PDF
        file_path = os.path.join(path, "event_direction_numbers.pdf")
        plt.savefig(file_path, format='pdf', bbox_inches='tight')

    return final_clusters_all_rounds, de_results_all_rounds





#  IDKfunction2.txt
#
#  IDK function
#
#  August 16, 2022
#
#  Can identify Intrinsic Dimension K of numeric matrix using 1 or more of the following methods:
#   MADA, CORRINT, MOM, PCAF0, TLE

import skdim
import numpy as np
import pandas as pd


# **** IDK function *******
#
#
def determine_nmf_ranks(PSIdf, *args, MADA=False, CORRINT=False, MOM=False, PCAFO=False, TLE=False, FS=False):
    #
    #   PSIdf is a pandas data frame with events (features) as columns, no missing values.
    #   event names are not needed; can be row names but not a column.
    #   Kmethods is a pandas series with True/False for each available method.
    #   Available methods are: MADA, CORRINT, MOM, PCAF0, TLE, FS.
    #
    # MADA is Manifold-Adaptive Dimension Estimation
    # CORRINT is Correlation Dimension
    # MOM is Method of Moments
    # PCAFO is the Fukunaga-Olsen version of the linear PCA method
    # TLE is Tight Local intrinsic dimensionality Estimator
    # FS is the FisherS method (PCA w/ l-descrim)

    #   Set up pandas series for estimates of K.  NaN for not done or error.
    Kests = {'MADA': np.nan, 'CORRINT': np.nan, 'MOM': np.nan, 'PCAFO': np.nan, 'TLE': np.nan, 'FS': np.nan}
    #
    Kests = pd.Series(Kests)
    #
    if MADA:
        mada = skdim.id.MADA().fit(PSIdf)
        Kests['MADA'] = mada.dimension_

    if CORRINT:
        corrint = skdim.id.CorrInt().fit(PSIdf)
        Kests['CORRINT'] = corrint.dimension_
    #
    if MOM:
        Mom = skdim.id.MOM().fit(PSIdf)
        Kests['MOM'] = Mom.dimension_
    #
    if PCAFO:
        pcafo = skdim.id.lPCA(ver='FO').fit(PSIdf)
        Kests['PCAFO'] = pcafo.dimension_
    #
    if TLE:
        tle = skdim.id.TLE(epsilon=0.0001).fit(PSIdf)
        Kests['TLE'] = tle.dimension_

    if FS:
        Fs = skdim.id.FisherS().fit(PSIdf)
        Kests['FS'] = Fs.dimension_
    #
    return Kests


import os
import pandas as pd
import numpy as np
from sklearn.metrics import f1_score

ground_truth_file = "/Users/tha8tf/MyProjects/Leucegene-positive-controls-F1.xlsx"
real_clusters = pd.read_excel(ground_truth_file, header=0, index_col=0)



main_folders = ["/Users/tha8tf/MyProjects/testsOncosplice/all_evals/all_defaults_rank_None_force_broad_off", "/Users/tha8tf/MyProjects/testsOncosplice/all_evals/all_defaults_rank_None_force_broad_on", "all_defaults_rank_30_force_broad_off", "all_defaults_rank_30_force_broad_on"]
# main_folders = ["force_broad_on_rank_30_efficient", "force_broad_on_rank_30_og_code"]
results = []  # To store F1 scores

for main_folder in main_folders:
    for round_folder in ["Round1", "Round2", "Round3"]:
        round_clusters_file = os.path.join(main_folder, round_folder, "round_clusters.txt")
        round_clusters = pd.read_csv(round_clusters_file, sep="\t", index_col=0)  # Adjust the separator if needed
        round_clusters = round_clusters.loc[real_clusters.index, :]

        round_results = []  # To store F1 scores for the current round

        for k in range(np.shape(real_clusters)[1]):
            for i in range(np.shape(round_clusters)[1]):
                y_true = real_clusters.iloc[:, k]
                y_pred = round_clusters.iloc[:, i]
                f1 = f1_score(y_true, y_pred)
                print(f"F1 Score for {real_clusters.columns.values[k]} vs. {round_clusters.columns.values[i]}: {f1}")

                round_results.append({
                    "Ground_Truth": real_clusters.columns.values[k],
                    "Generated_Cluster": round_clusters.columns.values[i],
                    "Round": round_folder,
                    "F1_Score": f1
                })

        results.append(round_results)


# Save the F1 scores to a CSV file in the main folder
output_file = "f1_scores_results.csv"
output_df = pd.DataFrame([result for round_result in results for result in round_result])
output_df.to_csv(os.path.join(main_folders[1], output_file), index=False)


import pandas as pd
from spectra.round_wrapper import *
from spectra.preprocess import *
from spectra.remove_redundancy import *
from spectra.feature_selection import *
from spectra.median_impute import *
from spectra.visualizations import *
import os
import argparse
import time
import matplotlib.pyplot as plt
import warnings
from matplotlib.backends.backend_pdf import PdfPages

plt.rcParams['pdf.fonttype'] = 42
plt.rcParams['font.family'] = 'Arial'


def oncosplice(args):
    psi_file_path = args.psi_file
    n_metadata_cols = args.n_metadata_cols
    corr_threshold = args.remove_redundant_events_threshold
    wd = args.wd
    dir = args.dir
    pca_corr_threshold = args.pca_threshold
    npcs = args.npcs
    rank_str = args.rank
    if rank_str == "k30":
        rank = 30
    else:
        rank = None
    print("NMF Rank used:")
    print(rank)
    rank_round1 = args.force_broad
    if rank_round1 == "on":
        rank1 = 2
    else:
        rank1 = rank
    print("Force broad used:")
    print(rank1)
    min_group_size = args.min_group_size
    dPSI = args.dpsi
    dPSI_p_val = args.dpsi_pval
    min_differential_events = args.min_differential_events
    top_n_differential_events = args.top_n_events
    conservation = args.conservation
    depletion_corr_threshold = args.depletion_corr
    speed = args.speed
    n_rounds = args.n_rounds
    fold_threshold = args.fold_threshold  # 0.3
    samples_differing = args.samples_differing  # 4
    corr_threshold_intercorr = args.corr_feature_selection  # 0.2
    corr_n_events = args.corr_n_events  # 10 or 5
    print("number of rounds to iterate through:")
    print(n_rounds)

    # read the psi file in and separate out the metadata columns from numeric PSI data
    formatted_psi_file, metadata = format_psi_matrix(psi_file_path, n=n_metadata_cols)
    metadata.index = formatted_psi_file.index
    formatted_psi_file = formatted_psi_file.dropna(how='all')
    metadata = metadata.loc[formatted_psi_file.index, :]

    st = time.time()
    metadata = variance_based_feature_selection(formatted_psi_file, metadata, fold_threshold=fold_threshold, samples_differing=samples_differing)
    et = time.time()
    el_t = (et-st) / 60
    print(f"finished variance based feature selection: {el_t} minutes")
    metadata = metadata.loc[metadata['high_variance'] == True, :]
    formatted_psi_file = formatted_psi_file.loc[metadata.index, :]

    if speed == "yes":
        st = time.time(); st_cpu = time.process_time()
        formatted_psi_file = fast_intercorrelation_based_feature_selection(formatted_psi_file, corr_threshold=corr_threshold_intercorr, corr_n_events=corr_n_events)
        et_cpu = time.process_time(); et = time.time()
        el_t = (et - st) / 60
        el_t_cpu = (et_cpu-st_cpu) / 60
        print(f"finished fast intercorrelated based feature selection: {el_t} minutes")
        print(f"finished fast intercorrelated based feature selection: {el_t_cpu} CPU minutes")
        
    else:
        st = time.time(); st_cpu = time.process_time()
        print("beginning intercorrelated based feature selection. This may take a while if input PSI file has more than 30,000 events (typical time 1 hour)...")
        formatted_psi_file = intercorrelation_based_feature_selection(formatted_psi_file, corr_threshold=corr_threshold_intercorr, corr_n_events=corr_n_events)
        et_cpu = time.process_time(); et = time.time()
        el_t = (et - st) / 60
        el_t_cpu = (et_cpu-st_cpu) / 60
        print(f"finished intercorrelated based feature selection: {el_t} minutes")
        print(f"finished fast intercorrelated based feature selection: {el_t_cpu} CPU minutes")

    metadata = metadata.loc[formatted_psi_file.index, :]

    # impute the psi matrix for certain downstream steps
    # formatted_psi_file_imp, metadata_imp = format_psi_matrix(imputed_psi_file_path)
    formatted_psi_file_imp = median_impute(formatted_psi_file)
    formatted_psi_file_imp.index = formatted_psi_file.index

    # Find the non-redundant splicing events
    st = time.time()
    list_of_events = remove_redundant_splicing_events(formatted_psi_file, corr_threshold, metadata)
    et = time.time()
    el_t = (et - st) / 60
    print(f"Finised redundant splicing event removal: {el_t} minutes")
    list_of_events = list(set(list_of_events))

    # Subset the PSI matrix to only the non-redundant events
    formatted_psi_file_0 = formatted_psi_file.loc[list_of_events, :]
    formatted_psi_file_imp_0 = formatted_psi_file_imp.loc[list_of_events, :]

    # Set working directory to save evaluations
    path = os.path.join(wd, dir)
    os.mkdir(path)
    os.chdir(path)

    # write the essential files out for future assessments
    list_of_events_df = pd.DataFrame(list_of_events)
    list_of_events_df.to_csv("list_of_non_redundant_events.txt", sep="\t")

    metadata.to_csv("final_metadata.txt", sep="\t")

    # Initialize an empty dictionaries to store DEG dataframes,
    final_clusters_dict = {}
    depleted_psi_file_after_round_imp_dict = {}
    depleted_psi_file_after_round_dict = {}
    deg_results_all_dict = {}

    print("Number of PCs used for feature selection")
    print(npcs)

    # feature selection prior to Round 1
    st = time.time()
    pca_events_round = get_events_from_pca(formatted_psi_file_imp, formatted_psi_file, list_of_events, corr_threshold=pca_corr_threshold, n_components=npcs)
    et = time.time()
    el_t = (et - st) / 60
    print(f"finished PCA based feature selection round 1: {el_t} minutes")

    pca_events_round_df = pd.DataFrame(pca_events_round)
    pca_events_round_df.to_csv("pca_events_round_1.txt", sep="\t")

    print("STARING ROUND 1...")
    print("Number of events prior to entering ROUND 1: ")
    print(len(formatted_psi_file_0.index))

    # Round 1 OncoSplice
    st = time.time()
    final_clusters_i, depleted_psi_file_after_round_imp_i, depleted_psi_file_after_round_i, deg_results_all_i = round_wrapper(
        filename="Round1", full_psi_file=formatted_psi_file_0, full_imputed_psi_file=formatted_psi_file_imp_0, metadata=metadata,
        highly_variable_events=pca_events_round, rank=rank1, min_group_size=min_group_size, dPSI=dPSI, dPSI_p_val=dPSI_p_val,
        min_differential_events=min_differential_events, top_n_differential_events=top_n_differential_events, conservation=conservation, strictness="tough",
        depletion_corr_threshold=depletion_corr_threshold, write_files=True, speed_corr="og")

    et = time.time()
    el_t = (et - st) / 60
    print(f"Time taken for round 1 clustering: {el_t} minutes")

    # Round i OncoSplice (tough)
    depleted_events_round_i = depleted_psi_file_after_round_i.index.to_list()

    print("Number of events removed after ROUND 1: ")
    print(len(formatted_psi_file_0.index) - len(depleted_psi_file_after_round_i.index))

    deg_results_all_i['cluster'] = 'R1_' + deg_results_all_i['cluster']
    # Add a meaninful prefix to all column names
    prefix = "R1_C"
    final_clusters_i = final_clusters_i.rename(columns=lambda x: prefix + str(x))

    # store the mandatory round 1 outputs in respective dictionaries
    key = "Round1"
    final_clusters_dict[key] = final_clusters_i
    # depleted_psi_file_after_round_imp_dict[key] = depleted_psi_file_after_round_imp_i
    # depleted_psi_file_after_round_dict[key] = depleted_psi_file_after_round_i
    deg_results_all_dict[key] = deg_results_all_i

    for round_j in range(2, n_rounds+1):

        # feature selection prior to Round J
        st = time.time()
        pca_events_round_j = get_events_from_pca(formatted_psi_file_imp, formatted_psi_file, depleted_events_round_i,
                                                 corr_threshold=pca_corr_threshold, n_components=npcs)

        et = time.time()
        el_t = (et - st) / 60
        print(f"finished PCA based feature selection round {round_j}: {el_t} minutes")

        pca_events_round_df_j = pd.DataFrame(pca_events_round_j)
        pca_events_round_df_j.to_csv(f"pca_events_round_{round_j}.txt", sep="\t")

        print(f"STARING ROUND {round_j}...")
        print(f"Number of events prior to entering ROUND {round_j}: ")
        print(len(depleted_psi_file_after_round_i.index))

        # Round j Oncosplice
        st = time.time()
        final_clusters_j, depleted_psi_file_after_round_imp_j, depleted_psi_file_after_round_j, deg_results_all_j = round_wrapper(
            filename=f"Round{round_j}", full_psi_file=depleted_psi_file_after_round_i,
            full_imputed_psi_file=depleted_psi_file_after_round_imp_i, highly_variable_events=pca_events_round_j, metadata=metadata, rank=rank,
            min_group_size=min_group_size, dPSI=dPSI, dPSI_p_val=dPSI_p_val, min_differential_events=min_differential_events,
            top_n_differential_events=top_n_differential_events, conservation=conservation, strictness="tough",
            depletion_corr_threshold=depletion_corr_threshold, write_files=True, speed_corr="og")

        et = time.time()
        el_t = (et - st) / 60
        print(f"Time taken for round {round_j}: {el_t} minutes")

        depleted_events_round_j = depleted_psi_file_after_round_j.index.to_list()
        print(f"Number of events removed after ROUND {round_j}: ")
        print(len(depleted_psi_file_after_round_i.index) - len(depleted_psi_file_after_round_j.index))

        deg_results_all_j['cluster'] = f'R{round_j}_' + deg_results_all_j['cluster']
        # Add a meaninful prefix to all column names
        prefix = f"R{round_j}_C"
        final_clusters_j = final_clusters_j.rename(columns=lambda x: prefix + str(x))

        # store the outputs in respective dictionaries
        key = f"Round{round_j}"
        final_clusters_dict[key] = final_clusters_j
        # depleted_psi_file_after_round_imp_dict[key] = depleted_psi_file_after_round_imp_j
        # depleted_psi_file_after_round_dict[key] = depleted_psi_file_after_round_j
        deg_results_all_dict[key] = deg_results_all_j

        # reset vars for next round
        depleted_events_round_i = depleted_events_round_j
        depleted_psi_file_after_round_i = depleted_psi_file_after_round_j
        depleted_psi_file_after_round_imp_i = depleted_psi_file_after_round_imp_j

        if len(depleted_psi_file_after_round_i) < top_n_differential_events:
            warnings.warn(f"Less than 50 splicing events pending after round {round_j}. Halting further iterative clustering process...")
            break

    path = os.path.join(path, "FinalResults")
    os.mkdir(path)
    os.chdir(path)

    final_clusters_all_rounds = pd.concat(final_clusters_dict.values(), axis=1)
    final_clusters_all_rounds.to_csv("MergedResults.txt", sep="\t")

    deg_results_all_rounds = pd.concat(deg_results_all_dict.values(), axis=0, ignore_index=True)

    # Grouping by cluster assignment and type of feature, and counting occurrences
    grouped_deg_df = deg_results_all_rounds.groupby(['cluster', 'EventAnnotation']).size().unstack(fill_value=0)
    grouped_deg_df.to_csv("event_annotations_numbers.txt", sep='\t')

    # Calculating percentage of each type of feature in each cluster
    grouped_annotation_percentage = grouped_deg_df.div(grouped_deg_df.sum(axis=1), axis=0) * 100
    grouped_annotation_percentage.to_csv("event_annotations_percentage.txt", sep='\t')

    # Define color palette for the types of features
    colors = sns.color_palette("Set1")

    # Plotting horizontal stacked bar graph
    plt.figure(figsize=(10, 6))
    grouped_annotation_percentage.plot(kind='barh', stacked=True, color=colors)
    plt.title('Percentage of Features in Each Type by Cluster Assignment')
    plt.ylabel('Cluster Assignment')
    plt.xlabel('Percentage')
    plt.legend(title='Event Annotation')
    plt.tight_layout()  # Adjust layout to fit cleanly in PDF
    file_path = os.path.join(path, "event_annotations_percentage.pdf")
    plt.savefig(file_path, format='pdf', bbox_inches='tight')

    plt.figure(figsize=(10, 6))
    grouped_deg_df.plot(kind='barh', stacked=True, color=colors)
    plt.title('Number of Features in Each Type by Cluster Assignment')
    plt.ylabel('Cluster Assignment')
    plt.xlabel('Number of Features')
    plt.legend(title='Event Annotation')
    plt.tight_layout()  # Adjust layout to fit cleanly in PDF
    file_path = os.path.join(path, "event_annotations_numbers.pdf")
    plt.savefig(file_path, format='pdf', bbox_inches='tight')

    # Grouping by cluster assignment and type of feature, and counting occurrences
    grouped_deg_df = deg_results_all_rounds.groupby(['cluster', 'event_direction'])['ClusterID'].nunique().unstack(fill_value=0)
    grouped_deg_df.to_csv("event_direction_numbers.txt", sep='\t')

    # Calculating percentage of each type of feature in each cluster
    grouped_annotation_percentage = grouped_deg_df.div(grouped_deg_df.sum(axis=1), axis=0) * 100
    grouped_annotation_percentage.to_csv("event_direction_percentage.txt", sep='\t')

    # Plotting horizontal stacked bar graph
    plt.figure(figsize=(10, 6))
    grouped_annotation_percentage.plot(kind='barh', stacked=True, color=colors)
    plt.title('Percentage of Cluster IDs in Each Type by Cluster/Group Assignment')
    plt.ylabel('Cluster Assignment')
    plt.xlabel('Percentage')
    plt.legend(title='Event Annotation')
    plt.tight_layout()  # Adjust layout to fit cleanly in PDF
    file_path = os.path.join(path, "event_direction_percentage.pdf")
    plt.savefig(file_path, format='pdf', bbox_inches='tight')

    plt.figure(figsize=(10, 6))
    grouped_deg_df.plot(kind='barh', stacked=True, color=colors)
    plt.title('Number of ClusterID in Each Type by Cluster Assignment')
    plt.ylabel('Cluster Assignment')
    plt.xlabel('Number of Features')
    plt.legend(title='Event Annotation')
    plt.tight_layout()  # Adjust layout to fit cleanly in PDF
    file_path = os.path.join(path, "event_direction_numbers.pdf")
    plt.savefig(file_path, format='pdf', bbox_inches='tight')

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Run Oncosplice on the PSI file')
    parser.add_argument('--wd', type=str, default=None, help='set working directory')  # dont change wd; instead have this as the path to write files to
    parser.add_argument('--dir', type=str, default=None, help='directory name to create for writing the results')  
    parser.add_argument('--psi_file', type=str, default=None, help='the path to the PSI file (with missing values)')
    parser.add_argument('--n_metadata_cols', type=int, default=11, help='expected number of metadata columns appended in front of PSI file')
    parser.add_argument('--remove_redundant_events_threshold', type=float, default=0.8, help='correlation threshold for removing redundant splicing events. default: 0.8')
    parser.add_argument('--pca_threshold', type=float, default=0.4, help='correlation threshold for doing PCA based feature selection. default: 0.4')
    parser.add_argument('--npcs', type=int, default=30, help='number of PCs to account for in PCA based feature selection. default: 30')
    parser.add_argument('--rank', type=str, default="k30", help='rank for NMF based clustering. use "None" for no rank. default: k30 (original Oncosplice algorithm)')
    parser.add_argument('--force_broad', type=str, default="off", help='Whether to find broad 2 clusters in Round 1 of Oncosplice. default: off')
    parser.add_argument('--min_group_size', type=int, default=5, help='minimum group size required for calculating differential events. default: 5')
    parser.add_argument('--dpsi', type=float, default=0.1, help='delta psi required to be considered as differential event. default: 0.1')
    parser.add_argument('--dpsi_pval', type=float, default=0.05, help='p value threshold for differential events. default: 0.05')
    parser.add_argument('--min_differential_events', type=int, default=100, help='minimum number of differential events required for a cluster to not be removed. default: 100')
    parser.add_argument('--top_n_events', type=int, default=150, help='top n differential events; max number of marker events per clusters for LinearSVM')  # add more details
    parser.add_argument('--conservation', type=str, default="stringent", help='strategy for cluster removal. see code comments for more details.')
    parser.add_argument('--depletion_corr', type=float, default=0.4, help='depletion correlation for keeping only the unique events after a round')
    parser.add_argument('--speed', type=str, default="og", help='if faster algorithm for correlation should be used. use "yes" for faster algorithm. default: og ')
    parser.add_argument('--n_rounds', type=int, default=3, help='number of rounds of iterative clustering. default: 3 sets')
    parser.add_argument('--fold_threshold', type=float, default=0.3,
                        help='fold threshold for variance based filtering. default: 0.3')
    parser.add_argument('--samples_differing', type=int, default=4,
                        help='n corresponding to n_highest - n_lowest in fold calculation in variance based filtering. default: 4')
    parser.add_argument('--corr_feature_selection', type=float, default=0.2,
                        help='intercorrelation feature selection correlation. default: 0.2')
    parser.add_argument('--corr_n_events', type=int, default=5,
                        help='minimum number of events a splicing event should be correlated with. default: 5')
    arg = parser.parse_args()
    start_time = time.time(); start_time_cpu = time.process_time()  # Record the start time
    oncosplice(arg)
    end_time_cpu = time.process_time(); end_time = time.time()  # Record the end time
    elapsed_time = end_time - start_time  # Calculate the elapsed time in seconds
    elapsed_minutes = elapsed_time / 60  # Convert elapsed time to minutes
    print(f"Code execution time: {elapsed_minutes:.5f} minutes")
    
    elapsed_time_cpu = end_time_cpu - start_time_cpu  # Calculate the elapsed time in seconds
    elapsed_minutes_cpu = elapsed_time_cpu / 60  # Convert elapsed time to minutes
    print(f"CPU Code execution time: {elapsed_minutes_cpu:.5f} minutes")



# Configuration file for the Sphinx documentation builder.

import os
import sys

# -- Path setup --------------------------------------------------------------

# Add the src/ directory to sys.path for autodoc to find 'spectra'
sys.path.insert(0, os.path.abspath('../spectra'))

# -- Project information -----------------------------------------------------

project = 'SPECTRA'
copyright = '2025, Kairavee Thakkar'
author = 'Kairavee Thakkar'
release = '0.1.0'

# -- General configuration ---------------------------------------------------

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx.ext.todo',
    'sphinx.ext.autosummary',
    'myst_parser',  # enable Markdown support; remove if not using Markdown
]

# Generate autosummary pages automatically
autosummary_generate = True

# Templates path
templates_path = ['_templates']

# Patterns to ignore when looking for source files
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# Enable TODOs in the documentation
todo_include_todos = True

# -- Options for HTML output -------------------------------------------------

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

# -- Napoleon settings for NumPy-style docstrings ----------------------------

napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True

# -- If using Markdown with myst_parser --------------------------------------

source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown',
}

# -- Autodoc options ---------------------------------------------------------

autodoc_member_order = 'bysource'
autodoc_inherit_docstrings = True



import seaborn as sns
import matplotlib.pyplot as plt
import os


def visualize_events_bars(deg_df, output_loc):

    # Grouping by cluster assignment and type of feature, and counting occurrences
    grouped_deg_df = deg_df.groupby(['cluster', 'EventAnnotation']).size().unstack(fill_value=0)

    # Calculating percentage of each type of feature in each cluster
    grouped_annotation_percentage = grouped_deg_df.div(grouped_deg_df.sum(axis=1), axis=0) * 100

    # Define color palette for the types of features
    colors = sns.color_palette("Set1")

    # Plotting horizontal stacked bar graph
    plt.figure(figsize=(10, 6))
    grouped_annotation_percentage.plot(kind='barh', stacked=True, color=colors)
    plt.title('Percentage of Features in Each Type by Cluster Assignment')
    plt.ylabel('Cluster Assignment')
    plt.xlabel('Percentage')
    plt.legend(title='Event Annotation')
    plt.tight_layout()  # Adjust layout to fit cleanly in PDF
    file_path = os.path.join(output_loc, "event_annotations_percentage.pdf")
    plt.savefig(file_path, format='pdf', bbox_inches='tight')
    plt.show()

    plt.figure(figsize=(10, 6))
    grouped_deg_df.plot(kind='barh', stacked=True, color=colors)
    plt.title('Number of Features in Each Type by Cluster Assignment')
    plt.ylabel('Cluster Assignment')
    plt.xlabel('Number of Features')
    plt.legend(title='Event Annotation')
    plt.tight_layout()  # Adjust layout to fit cleanly in PDF
    file_path = os.path.join(output_loc, "event_annotations_numbers.pdf")
    plt.savefig(file_path, format='pdf', bbox_inches='tight')
    plt.show()

    return grouped_deg_df, grouped_annotation_percentage


def psi_annotated_heatmap(psi_file, deg_df, clusters_df):

    heatmap_matrix = 2
    return heatmap_matrix


def binary_cluster_heatmap(clusters_df):

    heatmap_matrix = 2

    return heatmap_matrix




import pandas as pd
import numpy as np


def m(x, w):
    """Weighted Mean"""
    return np.sum(x * w) / np.sum(w)


def cov(x, y, w):
    """Weighted Covariance"""
    return np.sum(w * (x - m(x, w)) * (y - m(y, w))) / np.sum(w)


def weighted_corr(x, y, w):
    """Weighted Correlation"""
    return cov(x, y, w) / np.sqrt(cov(x, x, w) * cov(y, y, w))


def weighted_corr_vect(x, y, w):
    """Weighted Correlation"""
    numerator_cov = np.cov(x, y, aweights=w)
    numerator_cov = numerator_cov[0,0]
    denominator_cov1 = np.cov(x, x, aweights=w)
    denominator_cov1 = denominator_cov1[0,0]
    denominator_cov2 = np.cov(y, y, aweights=w)
    denominator_cov2 = denominator_cov2[0,0]

    return numerator_cov/np.sqrt(denominator_cov1 * denominator_cov2)


def calculate_sample_weights(full_psi_file):

    percentage_non_missing = (1 - full_psi_file.isna().mean()) * 100

    # The mean() method computes the mean of the boolean mask for each column, which gives the proportion of NaN values.

    weights = pd.DataFrame(percentage_non_missing, columns=['Percentage non-missing values'])

    return weights


def calculate_weighted_correlation(nmf_basis_matrix, full_psi_file, full_imputed_psi_file, weights, speed="vectorized"):

    nmf_basis_matrix = pd.DataFrame(nmf_basis_matrix).transpose()  # the shape of this matrix should be number of ranks by number of samples
    nmf_basis_matrix = nmf_basis_matrix.set_axis(full_psi_file.columns, axis=1)  # make sure the column names of the nmf matrix is the same as sample names (columns of psi file)

    weights = np.array(weights.loc[:, 'Percentage non-missing values'])

    corr_matrix = np.zeros([np.shape(nmf_basis_matrix)[0], np.shape(full_psi_file)[0]])  # create an empty array (of zeros) of size number of clusters by number of splicing events

    if speed == "vectorized":
        for cluster in range(np.shape(nmf_basis_matrix)[0]):
            for event in range(np.shape(full_imputed_psi_file)[0]):
                corr_matrix[cluster, event] = weighted_corr_vect(nmf_basis_matrix.iloc[cluster, :], full_imputed_psi_file.iloc[event, :], weights)
    else:
        for cluster in range(np.shape(nmf_basis_matrix)[0]):
            for event in range(np.shape(full_psi_file)[0]):
                corr_matrix[cluster, event] = weighted_corr(nmf_basis_matrix.iloc[cluster, :], full_psi_file.iloc[event, :], weights)  # this is a slow step -- how can you optimize this?

    return corr_matrix


def deplete_events(nmf_basis_matrix, full_psi_file, full_imputed_psi_file, corr_threshold=0.3, strictness="tough", speed="vectorized"):

    weights = calculate_sample_weights(full_psi_file)

    corr_matrix = calculate_weighted_correlation(nmf_basis_matrix, full_psi_file, full_imputed_psi_file, weights, speed)

    corr_matrix = np.array(corr_matrix)  # this corr matrix represents the correlation between a cluster and splicing event (based on samples/patients)

    mask = abs(corr_matrix) < corr_threshold

    if strictness == "tough":
        has_low_corr = np.all(mask, axis=0)
    else:
        has_low_corr = np.any(mask, axis=0)  # should this check for the TRUE in at least one cluster or in ALL clusters?

    depleted_psi_file = full_psi_file.loc[has_low_corr]

    return depleted_psi_file



import pandas as pd
import numpy as np
from sklearn.svm import LinearSVC


def generate_train_data(psi_file_with_diff_events, nmf_binarized_output):

    # for each of the nmf clusters provided, compute the centroid for each cluster for the differential splicing events using the output_file arg (provided in the output file?)

    centroids = pd.DataFrame()
    for col in range(nmf_binarized_output.shape[1]):

        col_all_samples = nmf_binarized_output[:, col]
        # col_all_samples = col_all_samples.to_numpy()
        col_cluster_samples = np.where(col_all_samples == 1)  # this is a tuple which is ridiculously annoying. it looks like: (array([ 63,  65,  71, 234, 349, 364]),) so need I am extracting "first" element in the next line to compute centroid to get only the array which is indices.

        centroid_vec = psi_file_with_diff_events.iloc[:, col_cluster_samples[0]].mean(axis=1)  # row means (rows are events) for each group
        centroids[col] = centroid_vec
                    
    return centroids


def classify(train, imputed_psi_file_with_diff_events, groups, conservation="stringent"):

    # Y is median imputed psi file and X is "grplst"/groups in the original function by MV
    # regression fit on the train dataset
    regression = LinearSVC()  # alternative is OneVsRestClassifier function in sklearn -- why not this one?
    regression.fit(train.transpose(), groups)  # train should be # of ranks by # of differential splicing events ("samples" as per python documentation are the ranks/clusters) and "features" are splicing events and groups is # of ranks/clusters by # of patients/samples
    # in the previous regression fit, why do we choose just the first column (it refers to the first patient/sample only, so what does the reproducibility looks like here?
    # i checked this by selecting a column or a patient/sample that has exactly 1 cluster assignment when k =2 and the decision function values are not identical but they are similar so yes, there is some issues regarding reproducibility here. here is an example from 3 different columns for the groups[:,"mysterious column"]

    # prob[1:10]
    # Out[287]:
    # array([-0.21518692, 0.1017927, 0.00177943, 0.89216488, 0.2264813,
    #        1.24986209, 0.97950145, 1.44557026, 1.15520506])
    # prob4[1:10]
    # Out[288]:
    # array([-2.15928144e-01, 1.01157955e-01, 1.12518304e-03, 8.91793519e-01,
    #        2.25907636e-01, 1.24954443e+00, 9.79122635e-01, 1.44533371e+00,
    #        1.15487401e+00])
    # prob10[1:10]
    # Out[289]:
    # array([-0.21470678, 0.10214063, 0.00216086, 0.89218438, 0.22676602,
    #        1.24977117, 0.979506, 1.44538875, 1.15514289])

    prob = regression.decision_function(imputed_psi_file_with_diff_events.transpose())  # in the PSI file here, it should be a # of samples/patients by # of features format so we are transposing it

    if len(groups) > 2:
        final_clusters = prob > 0
        #print(prob)
    else:  # this is when NMF rank = 2;
        # if number of groups/rank=2 then choose the cluster for which the decision function score is greater than 0.5 or less than -0.5 ("stringent") (> 0.25 and < -0.25 in case of "conservative"/"expanded")
        if conservation == "stringent":
            decision_score_threshold = 0.5
        else:
            decision_score_threshold = 0.25
        final_cluster_1 = prob > decision_score_threshold
        final_cluster_2 = prob < -decision_score_threshold
        final_clusters = np.stack([final_cluster_1, final_cluster_2])
        final_clusters = final_clusters.transpose()

    final_clusters = final_clusters.astype(int)
    final_clusters = pd.DataFrame(final_clusters, index=imputed_psi_file_with_diff_events.columns)

    return final_clusters



import pandas as pd
import numpy as np
import scipy.stats
import statsmodels.stats.multitest as ssm
import os


# ### Define function to perform some data checks before mwu computation

def mwuChecks(dataf, grpdf, grpvar='grp', min_group_size=6):
    #
    #   This function performs some checks before calling mwuCompute().
    #
    #   dataf is a pandas data frame with samples as columns and events (features) as rows.
    #   Event names must be the row names (index), not in a column.
    #   Missing values (np.nan='NaN') are allowed
    #
    #   grpdf is a pandas data frame with a group membership column ('grp' is the default ).
    #   Sample names must be the row names (index), not in a column.
    #   Aside from ordering, must be 1-1 match between dataf.columns and grpdf.index
    #
    #   min_group_size is minimum group size for kw test to be computed.
    #   min_group_size = 6 is based on statistical rule-of-thumb for Mann-Whitney 2-group test.
    #
    #  Returns integer to indicate course of action:
    #     0 = Problem, do not run mwuCompute
    #     1 = Run mwuCompute

    #    print(grpvar)
    #    print(type(grpvar))

    # --- Check for 1-1 match between sample names provided in grpdf and the data matrix

    if set(dataf.columns) != set(grpdf.index):
        print('Group information is inconsistent with names in data matrix')
        return 0

    grpCounts = grpdf[grpvar].value_counts().to_frame()
    # returns a pandas df of counts, indexed by grp, decreasing by count
    grpCounts.columns = ['grpRawN']
    nGrps = grpCounts.shape[0]
    minGrpN = min(grpCounts['grpRawN'])
    # print('nGrps, minGrpN',nGrps,minGrpN)

    # -- Handle groups <> 2 --

    if nGrps < 2:
        print('Number of sample groups is < 2; Mann-Whitney test not conducted')
        return 0

    if nGrps > 2:
        print('Number of sample groups is > 2; Mann-Whitney test not conducted')
        return 2

    # -- Don't proceed if already know a group size is < minimum --

    if minGrpN < min_group_size:
        print('Mann-Whitney test not conducted: A group has fewer samples than minimum group size: ', minGrpN, ' < ',
              min_group_size)
        return 0

    return 1


# ### Define function to compute the mwu test and collect summary stats (Ns, medians)

def mwuCompute(dataf, grpdf, metadata_df, grpvar='grp', min_group_size=6):
    #
    #   This function performs the Mann-Whitney NP (rank) two independent groups test, aka Wilcoxon 2 grp indep test.
    #
    #   dataf is a pandas data frame with samples as columns and events (features) as rows.
    #   Event names must be the row names (index), not in a column.
    #   Missing values (np.nan='NaN') are allowed
    #
    #   grpdf is a pandas data frame with a group membership column.
    #   Sample names must be the row names (index), not in a column.
    #   Aside from ordering, must be 1-1 match between dataf.columns and grpdf.index
    #
    #   min_group_size is minimum group size for mwu test to be computed.
    #   6 is based on statistical rule-of-thumb for Mann-Whitney 2-group test.
    #
    #   Direction of comparison is by sorted grp levels, e.g. A vs. B, 1 vs 2.
    #   Sign of result is determined by comparing test statistic value to the null mean (not comparing group medians or means)
    #
    #   Returns pandas dataframe with columns containing group Ns (xNaN), group medians and means, value of mwu test statistic, sign and p-value.
    #   Returned df has row for each event, even if test was not computed (in this case mwu values will be NaN)

    groups = dataf.T.groupby(grpdf[grpvar])
    #    print(groups.groups)

    grpCounts = grpdf[grpvar].value_counts().to_frame()
    #    display(grpCounts)

    grpLevels = grpCounts.index
    grpLevels = grpLevels.tolist()
    #    print(grpLevels)
    SgrpLevels = sorted(grpLevels)
    #    print("SgrpLevels",SgrpLevels)

    g1Level = SgrpLevels[0]
    g2Level = SgrpLevels[1]
    #    print('g1Level',g1Level)
    #    print('g2Level',g2Level)

    g1 = groups.get_group(g1Level)
    g2 = groups.get_group(g2Level)
    #    print('g1 dat',g1)
    #    print('g2 dat',g2)
    #   count() method counts number of non-missing entries
    eventsToRun = ((g1.count() >= min_group_size) & (g2.count() >= min_group_size)).values
    #    print('eventsToRun, sum',eventsToRun,sum(eventsToRun))
    #    print(g1.iloc[:,eventsToRun])
    #    print(g2.iloc[:,eventsToRun])

    #   Set up df for results
    # tempdf.columns=[('N_' + str(gID)),('Median_' + str(gID))]

    result = pd.DataFrame(
        {
            ('N_' + str(g1Level)): g1.count(),
            ('N_' + str(g2Level)): g2.count(),
            ('Median_' + str(g1Level)): g1.median(),
            ('Median_' + str(g2Level)): g2.median(),
            ('Mean_' + str(g1Level)): g1.mean(),
            ('Mean_' + str(g2Level)): g2.mean(),
            'mwuStat': np.nan,
            'mwuPval': np.nan,
            'mwuAdjPval': np.nan,
            'mwuSign': np.nan,
            'Abs_Mean_Diff': np.nan,
        },
        index=dataf.index)

    #    display(result)

    #   Handle situation of no events meeting min_group_size criterion

    if sum(eventsToRun) == 0:
        print('Mann-Whitney test not conducted: No events meet minimum group size criterion of ', min_group_size, '.')
        return result

    #   Compute the mwu test statistic & p-value for eventsToRun

    mwu = scipy.stats.mannwhitneyu(
        g1.iloc[:, eventsToRun].values,
        g2.iloc[:, eventsToRun].values,
        alternative="two-sided",
        method="asymptotic",
        use_continuity=True,
        nan_policy='omit'
    )
    #    print(mwu)

    #   Populate stat & p-value columns
    result.loc[eventsToRun, "mwuStat"] = mwu.statistic
    result.loc[eventsToRun, "mwuPval"] = mwu.pvalue
    fdr_bh_test = ssm.multipletests(result.loc[eventsToRun, "mwuPval"].to_numpy(), method="fdr_bh")
    result.loc[eventsToRun, "mwuAdjPval"] = fdr_bh_test[1]

    #    display(result)

    #   Determine sign (direction) of test result based on expected mean of test stat under null
    nullMean = (result[('N_' + str(g1Level))] * result[('N_' + str(g2Level))]) / 2.0 + 0.5
    #    print(nullMean)
    result["mwuSign"] = np.select(
        [
            (result["mwuStat"] < nullMean),
            (result["mwuStat"] == nullMean),
            (result["mwuStat"] > nullMean),
        ],
        [-1, 0, 1],
        default=np.nan)

    #    display(result)

    result['cluster'] = 'c' + grpvar
    metadata_df = metadata_df.loc[result.index, ]
    result = pd.concat([metadata_df, result], axis=1)

    return result


def find_top_differential_events(exp_file, groups_file, min_group_size, metadata, output_loc, dPSI=0.1, dPSI_p_val=0.05, min_differential_events=100, top_n=150):

    all_reg_events = []
    remove_clusters = []

    groups_file_cols = groups_file.columns
    deg_cols = pd.Index(['N_cluster',
                         'N_others',
                         'Median_cluster',
                         'Median_others',
                         'Mean_cluster',
                         'Mean_others',
                         'mwuStat',
                         'mwuPval',
                         'mwuAdjPval',
                         'mwuSign',
                         'Abs_Mean_Diff',
                         'cluster'])
    metadata_cols = metadata.columns
    results_cols = metadata_cols.union(deg_cols, sort=False)

    results_all = pd.DataFrame(columns=results_cols)

    for cluster in range(len(groups_file_cols)):
        results_c0 = mwuCompute(exp_file, groups_file, grpvar=groups_file_cols[cluster], metadata_df=metadata, min_group_size=min_group_size)
        # results_c0 = results_c0[results_c0['mwuPval'] < pval_threshold]  # for using raw p value to determining significant events
        results_c0.loc[:, "Abs_Mean_Diff"] = abs(results_c0['Mean_0'] - results_c0['Mean_1'])
        results_c0.columns = results_cols
        results_c0 = calculate_coordinate_lengths(results_c0)
        results_c0 = determine_event_direction(results_c0)
        # filter based on deg requirements
        results_c0 = results_c0[results_c0['mwuAdjPval'] < dPSI_p_val]
        results_c0 = results_c0[results_c0["Abs_Mean_Diff"] > dPSI]
        if results_c0.shape[0] < min_differential_events:
            print(("Less than " + str(min_differential_events) + " significant diff events for Cluster " + groups_file_cols[cluster]))
            remove_clusters.append(cluster)
        else:
            # results_c0 = results_c0.sort_values(by=['Abs_Mean_Diff'], ascending=False)  # for using top n events sorted by fold change instead of p val
            results_c0 = results_c0.sort_values(by=['mwuAdjPval'])  # does not matter much whether you sort by raw or adjusted p value here -- results will be identical most likely
            results_all = pd.concat([results_all, results_c0], axis=0)  # equivalent to rbind
            file_path = os.path.join(output_loc, 'c_' + groups_file_cols[cluster] + "_vs_others_splicing_events.txt")
            results_c0.to_csv(file_path, sep="\t")
            results_c0 = results_c0.iloc[:top_n, :]
            all_reg_events = all_reg_events + results_c0.index.to_list()
            # top150_reg_events_file = "/Users/tha8tf/Documents/testsOncosplice/differential_events_per_cluster/all_reg_events_force_broad_nmf_cluster_adjp_0.05_" + cluster + ".txt"
            # results_c0.to_csv(top150_reg_events_file, sep='\t')

    all_reg_events_unique = pd.Series(all_reg_events).drop_duplicates()
    # print(("Number of unique differential events found for this round is " + str(len(all_reg_events_unique.tolist()))))

    remove_clusters = list(map(int, remove_clusters))  # converting string to integer for the clusters to remove so I can use these integers for filtering the NMF binary groups file

    return all_reg_events_unique, remove_clusters, results_all


def calculate_coordinate_lengths(deg_result):

    # Function to split the string by "|" or ":"
    def split_string(entry, sep):
        return entry.split(sep)

    # Apply the function to split the coordinates into numerator and denominator junctions (| is the separator)
    split_data = deg_result['Coordinates'].apply(split_string, sep="|")

    # Assign the split values to new columns
    deg_result['num_coordinates'] = split_data.apply(lambda x: x[0])
    deg_result['den_coordinates'] = split_data.apply(lambda x: x[1])

    num_coordinates = pd.DataFrame(deg_result['num_coordinates'].apply(split_string, sep=":"))
    num_coordinates['num_coordinates'] = num_coordinates['num_coordinates'].apply(lambda x: x[1])
    num_coordinates['num_coordinates'] = num_coordinates['num_coordinates'].apply(split_string, sep="-")
    num_coordinates['Coordinate1'] = pd.to_numeric(num_coordinates['num_coordinates'].apply(lambda x: x[0]))
    num_coordinates['Coordinate2'] = pd.to_numeric(num_coordinates['num_coordinates'].apply(lambda x: x[1]))
    num_coordinates['num_coordinate_diff'] = abs(num_coordinates['Coordinate1'] - num_coordinates['Coordinate2'])

    den_coordinates = pd.DataFrame(deg_result['den_coordinates'].apply(split_string, sep=":"))
    den_coordinates['den_coordinates'] = den_coordinates['den_coordinates'].apply(lambda x: x[1])
    den_coordinates['den_coordinates'] = den_coordinates['den_coordinates'].apply(split_string, sep="-")
    den_coordinates['Coordinate1'] = pd.to_numeric(den_coordinates['den_coordinates'].apply(lambda x: x[0]))
    den_coordinates['Coordinate2'] = pd.to_numeric(den_coordinates['den_coordinates'].apply(lambda x: x[1]))
    den_coordinates['den_coordinate_diff'] = abs(den_coordinates['Coordinate1'] - den_coordinates['Coordinate2'])

    deg_result['num_coordinate_diff'] = num_coordinates['num_coordinate_diff']
    deg_result['den_coordinate_diff'] = den_coordinates['den_coordinate_diff']

    return deg_result


def determine_event_direction(deg_result):

    # Junction1 | Junction2
    # Example: chr3: 129575563 - 129575562 | chr3:129575766 - 129575562
    # A = abs(129575563 - 129575562) = 1
    # B = abs(129575766 - 129575562) = 204
    # A = num_coordinate_diff
    # B = den_coordinate_diff
    # If A < B and dPSI > 0: type = inclusion
    # If A > B and dPSI < 0: type = inclusion
    # If A < B and dPSI < 0: type = exclusion
    # If A > B and dPSI > 0: type = exclusion

    deg_result['event_direction'] = 'inclusion'

    mask_junction = deg_result['num_coordinate_diff'] < deg_result['den_coordinate_diff']
    mask_regulation = deg_result['mwuSign'] < 0

    deg_result.loc[mask_junction & mask_regulation, 'event_direction'] = 'exclusion'
    deg_result.loc[~mask_junction & ~mask_regulation, 'event_direction'] = 'exclusion'

    deg_result.drop(columns=['num_coordinates', 'den_coordinates'], inplace=True)

    return deg_result




import pandas as pd
import numpy as np

# Variance based feature selection
def variance_based_feature_selection(formatted_psi_file, metadata, fold_threshold=0.2, samples_differing=3):

    metadata['high_variance'] = True

    for feature in formatted_psi_file.index:
        feature_vector = formatted_psi_file.loc[feature, :]
        feature_vector = feature_vector.dropna()
        feature_vector = feature_vector.sort_values(ascending=True)

        n_lowest = feature_vector.head(samples_differing).iloc[samples_differing-1]
        n_highest = feature_vector.tail(samples_differing).iloc[0]
        fold = n_highest - n_lowest
        metadata.loc[feature, 'high_variance'] = fold > fold_threshold

    return metadata


# Inter correlation step do this step if only less than 8000 to start with -- cut off for this 0.4 -- should bring it down to 1500
def intercorrelation_based_feature_selection(df, corr_threshold=0.5, corr_n_events=10):

    # Calculate correlation matrix
    df = df.transpose()

    corr_matrix = df.corr(method="pearson")

    # Get columns with correlation > 0.4
    correlated_features = (corr_matrix > corr_threshold).sum() > corr_n_events

    # Filter dataframe based on correlated features
    filtered_df = df[correlated_features.index[correlated_features]]

    filtered_df = filtered_df.transpose()

    return filtered_df


def fast_intercorrelation_based_feature_selection(df, corr_threshold=0.5, corr_n_events=10):
    # Calculate correlation matrix
    df = df.transpose()

    corr_matrix = fast_corr_matrix(data_matrix=df.to_numpy())

    corr_matrix = pd.DataFrame(corr_matrix, index=df.columns, columns=df.columns)

    # Get columns with correlation > 0.4
    correlated_features = (corr_matrix > corr_threshold).sum() > corr_n_events

    # Filter dataframe based on correlated features
    filtered_df = df[correlated_features.index[correlated_features]]

    filtered_df = filtered_df.transpose()

    return filtered_df


def test_func(df1, df2):
    norm1 = df1 - df1.mean(axis=0)
    norm2 = df2 - df2.mean(axis=0)
    sqsum1 = (norm1**2).sum(axis=0)
    sqsum2 = (norm2**2).sum(axis=0)
    return((norm1.T @ norm2) / np.sqrt(sqsum1.apply(lambda x: x*sqsum2)))


import numpy as np


def fast_corr_matrix(data_matrix):

    # Mask out NaNs by replacing them with the column mean, if necessary
    valid_data = np.nan_to_num(data_matrix, nan=np.nanmean(data_matrix, axis=0))

    # Subtract the mean from each column (centering the data)
    centered_data = valid_data - np.nanmean(data_matrix, axis=0)

    # Calculate the dot product of the centered data matrix with itself
    dot_product = np.dot(centered_data.T, centered_data)

    # Calculate the sum of squared deviations (like variance)
    sum_squared = np.sqrt(np.sum(centered_data ** 2, axis=0))

    # Outer product of sum_squared to get all combinations of feature variances
    denom = np.outer(sum_squared, sum_squared)

    # Calculate the correlation matrix
    corr_matrix = dot_product / denom

    # Set the diagonal to 1 (perfect correlation with self)
    np.fill_diagonal(corr_matrix, 1)

    return corr_matrix




import pandas as pd
import os
from .determine_rank import *
from .run_nmf import *
import time
from .metadata_analysis import *
from .linear_svm import *
from .correlation_depletion import *
from .visualizations import *


def round_wrapper(filename, full_psi_file, full_imputed_psi_file, highly_variable_events, metadata, rank=None, min_group_size=5,
                  dPSI=0.1, dPSI_p_val=0.05, min_differential_events=100, top_n_differential_events=150,
                  conservation="stringent", strictness="easy", depletion_corr_threshold=0.3, write_files=True,
                  speed_corr="og"):
    # a wrapper function that does for ONE iteration 1) nmf analysis 2) finds unique differential splicing events per NMF component and 3) linear SVC and 4) depletes the events for next round
    # full_psi_file: splicing events by samples dataframe of all splicing events (events after 1) 75% presence of samples  and 2) removing redundant events using inter-correlation analysis)
    # full_imputed_file: splicing events by samples dataframe of all splicing events (events with 75% presence of samples) but missing values imputed with median values for each splicing events
    # highly_variable_events: list of splicing events that are a subset of the events in the full psi file/full imputed file (usually events after feature selection step)
    # rank: optional integer value for the number of clusters for NMF analysis. alternatively, if user wants to detect automatically then, ignore.
    # min_group_size: integer value indicating the minimum number of samples that should belong to each cluster for valid differential expression analysis for the events
    # dPSI: fold change threshold. if events have a dPSI above this value and are significant based on dPSI_p_val value, then they are considered as differentially expressed events for a given cluster
    # dPSI_p_val: adjusted p value to use as reference to determine differential events
    # min_differential_events: integer value. if a cluster has less than this number of differential events
    # conservation: "stringent" or "expanded" to determine decision function threshold in linearSVM classification when NMF rank=2 or # of valid clusters = 2.
    # depletion_corr_threshold: correlation threshold to use to determine which events to deplete at the end of the round
    # write_files: writes out all the relevant files in txt form if set to True. otherwise, saves time by not writing/overwriting files.

    if write_files is True:
        os.mkdir(filename)
        cd = os.getcwd()

    # filter the full psi files (one with na and one with median imputed) with the highly variable events (currently comes from the file PCAbasedFeatureSelection.py ran before this function)
    formatted_psi_file_hve = full_psi_file.loc[highly_variable_events, :]
    formatted_psi_file_imp_hve = full_imputed_psi_file.loc[highly_variable_events, :]

    # when rank is not 2 or not provided by the user (that they trust more), then automatically calculate number of clusters for NMF to determine (aka rank)
    # requires the determineNMFRank.py script
    if rank is None:
        k_results = determine_nmf_ranks(PSIdf=formatted_psi_file_imp_hve.transpose(), MADA=True, CORRINT=True, MOM=True,
                                        PCAFO=True, TLE=True, FS=True)
        est_k = round(pd.Series.median(k_results, skipna=True))
        rank = est_k

    # run and time the NMF analysis using the runNMF.py script
    start_time = time.time()
    print(("Running NMF analysis with rank set to " + str(rank) + "..."))
    basis_matrix, binarized_output_test, refined_binarized_output = run_nmf(imputed_psi_file=formatted_psi_file_imp_hve,
                                                                            rank=rank)
    elapsed_time = (time.time() - start_time) / 60  # Convert seconds to minutes
    print("--- %.3f minutes ---" % elapsed_time)

    # for metadata analysis, which determines the differential splicing events, format the binary groups file (an output from NMF analysis) to match the expectations of the mwuCompute function
    groups_file = pd.DataFrame(
        refined_binarized_output.transpose())  # dimensions should be number of patients/samples (rows) by number of clusters (cols)
    groups_file.index = formatted_psi_file_hve.columns
    groups_file.columns = groups_file.columns.map(str)
    groups_file_cols = groups_file.columns

    if write_files is True:
        initial_clusters_path = os.path.join(cd, filename, "initial_round_clusters.txt")
        groups_file.to_csv(initial_clusters_path, sep="\t")

    remove_cols_from_nmf_clusters = []
    # check whether the data being provided is in good format and matches the sanity requirements for determining differential splicing events
    for col in range(len(groups_file_cols)):
        # print(col)
        mwu_check = mwuChecks(full_psi_file, groups_file, grpvar=groups_file_cols[col], min_group_size=min_group_size)
        # print(mwu_check)
        if mwu_check == 0:
            # print(col)
            # print(groups_file_cols[col])
            print(("differential expression not compatible for cluster " + groups_file_cols[col]))
            groups_file = groups_file.drop(groups_file_cols[col], axis='columns')
            # print(groups_file.columns)
            # refined_binarized_output = np.delete(refined_binarized_output, col, axis=0)
            remove_cols_from_nmf_clusters.append(col)

    refined_binarized_output = np.delete(refined_binarized_output, remove_cols_from_nmf_clusters, axis=0)

    start_time = time.time()
    print(("Finding top " + str(top_n_differential_events) + " differential splicing events per NMF cluster..."))
    diff_events, remove_clusters, results_all = find_top_differential_events(exp_file=full_psi_file, groups_file=groups_file,
                                                                             metadata=metadata,
                                                                             output_loc=os.path.join(cd, filename),
                                                                             min_group_size=min_group_size, dPSI=dPSI,
                                                                             dPSI_p_val=dPSI_p_val,
                                                                             min_differential_events=min_differential_events,
                                                                             top_n=top_n_differential_events)
    elapsed_time = (time.time() - start_time) / 60  # Convert seconds to minutes
    print("--- %.3f minutes ---" % elapsed_time)

    start_time = time.time()
    print(("Determined " + str(len(diff_events)) + " unique differential events across all clusters"))
    psi_file_with_diff_events = full_psi_file.loc[diff_events,]
    psi_file_with_diff_events_imp = full_imputed_psi_file.loc[diff_events,]
    elapsed_time = (time.time() - start_time) / 60  # Convert seconds to minutes
    print("--- %.3f minutes ---" % elapsed_time)

    refined_binarized_output_f = np.delete(refined_binarized_output, remove_clusters,
                                           axis=0)  # remove clusters (rows) with number of differential splicing events less than threshold (100)

    # Linear SVM/SVC training dataset generation by creating centroids for the NMF clusters

    centroids = generate_train_data(psi_file_with_diff_events=psi_file_with_diff_events,
                                    nmf_binarized_output=refined_binarized_output_f.transpose())  # splicing events by clusters matrix
    centroids = centroids.dropna(axis=0)
    psi_file_with_diff_events_imp = psi_file_with_diff_events_imp.loc[centroids.index, :]

    start_time = time.time()
    print(("Running Linear SVM on final " + str(np.shape(refined_binarized_output_f)[0]) + " clusters with..."))
    final_clusters = classify(train=centroids, imputed_psi_file_with_diff_events=psi_file_with_diff_events_imp,
                              groups=np.arange(np.shape(refined_binarized_output_f)[0]),
                              conservation=conservation)  # need to figure out a way to give more meaningful cluster name instead of an integer after this step (such as C1, etc etc).
    elapsed_time = (time.time() - start_time) / 60  # Convert seconds to minutes
    print("--- %.3f minutes ---" % elapsed_time)

    print(("Depleting events using a correlation threshold of " + str(depletion_corr_threshold) + "..."))
    if speed_corr == "og":
        start_time = time.time()
        depleted_psi_file_after_round = deplete_events(nmf_basis_matrix=basis_matrix, full_psi_file=full_psi_file,
                                                       full_imputed_psi_file=full_imputed_psi_file, corr_threshold=depletion_corr_threshold, strictness=strictness, speed="og")
        elapsed_time = (time.time() - start_time) / 60  # Convert seconds to minutes
        print("--- %.3f minutes ---" % elapsed_time)
    elif speed_corr == "efficient":
        start_time = time.time()
        depleted_psi_file_after_round = deplete_events(nmf_basis_matrix=basis_matrix, full_psi_file=full_psi_file,
                                                       full_imputed_psi_file=full_imputed_psi_file, corr_threshold=depletion_corr_threshold, strictness=strictness, speed="vectorized")
        elapsed_time = (time.time() - start_time) / 60  # Convert seconds to minutes
        print("--- %.3f minutes ---" % elapsed_time)

    depleted_psi_file_after_round_imp = full_imputed_psi_file.filter(depleted_psi_file_after_round.index, axis="rows")
    depleted_psi_file_after_round = full_psi_file.filter(depleted_psi_file_after_round.index, axis="rows")

    if write_files is True:
        diff_events_path = os.path.join(cd, filename, "input_linear_svm_features.txt")
        final_clusters_path = os.path.join(cd, filename, "round_clusters.txt")
        diff_events.to_csv(diff_events_path, sep="\t")
        final_clusters.to_csv(final_clusters_path, sep="\t")

    return final_clusters, depleted_psi_file_after_round_imp, depleted_psi_file_after_round, results_all



import pandas as pd
import os

# Define the path to your directory containing the folders
directory_path = '/data/salomonis2/LabFiles/Kairavee/oncosplice-KT/AltAnalyze3-Oncosplice/clean_evals/runs_version2_7k_events/'

# Define the order of mutations
mutation_order = ['MLL_fusions', 'U2AF1-S34_variants', 'CBFB-MYH11_fusions', 'PML-RARA_fusions', 'U2AF1-Q157_variants', 'ZRSR2_variants', 'RUNX1_fusions', 'SF3B1_variants', 'SRSF2-P95-Deletion', "no_mutation"]  # Provide your specific order

# Initialize an empty dictionary to store max F1 scores for each mutation
max_f1_scores = {}

all_folders = os.listdir(directory_path)
all_folders.remove('.DS_Store')

# Iterate over each folder in the directory
for folder in all_folders:
    folder_path = os.path.join(directory_path, folder)

    # Check if the current item is a directory
    if os.path.isdir(folder_path):
        file_path = os.path.join(folder_path, "f1_scores_results_nomutations_removed1.csv")

        # Check if the file "f1_scores.csv" exists in the current folder
        if os.path.exists(file_path):
            # Read the CSV file into a DataFrame
            df = pd.read_csv(file_path)

            # Iterate over each unique mutation in the first column
            for mutation in df['Ground_Truth'].unique():
                # Find the maximum F1 score for the current mutation
                max_f1 = df[df['Ground_Truth'] == mutation]['F1_Score'].max()

                # Update the dictionary with the maximum F1 score for the current mutation
                max_f1_scores[mutation] = max_f1

# Create a DataFrame to store the final results
    final_df = pd.DataFrame(columns=['mutation', 'max_f1'])
    df_path = os.path.join(folder_path, "final_max_f1_scores_no_mutations.csv")
# Populate the DataFrame with the mutations and their corresponding maximum F1 values
    for mutation in mutation_order:
        if mutation in max_f1_scores:
            final_df = final_df.append({'mutation': mutation, 'max_f1': max_f1_scores[mutation]}, ignore_index=True)
            # Write the DataFrame to a CSV file named "final_max_f1_scores.csv"
            final_df.to_csv(df_path, index=False)




import pandas as pd
import numpy as np


def format_psi_matrix(psi_file_path, n=11):

    psi_file = pd.read_csv(psi_file_path, sep="\t", header=0, index_col="UID")
    metadata = psi_file.iloc[:, 0:n]  # this needs to change maybe because not all PSI files have exact 11 columns of metadata
    formatted_psi_file = psi_file.iloc[:, n:]

    return formatted_psi_file, metadata


def remove_redundant_splicing_events(formatted_psi_file, corr_threshold, metadata):
    # print("corr.threshold for corr level being tested: " + str(corr_threshold))

    # create an empty list which will be appended with every gene run
    filtered_rows = []

    # get the list of genes
    rows_withonlygene = metadata.iloc[:, 0]
    rows_withonlygene = rows_withonlygene + ":"
    rows_withonlygene_np = rows_withonlygene.to_numpy()
    gene_list = pd.unique(rows_withonlygene)

    mat = formatted_psi_file

    for i in gene_list:
        # print(i)
        logical_vec = np.isin(rows_withonlygene_np, i)  # determine which are the rows in the PSI matrix associated with that gene
        gene_i_matrix = mat.iloc[logical_vec, ]  # get the events associated with the gene from PSI matrix
        gene_i_matrix = gene_i_matrix.transpose()  # transpose the filtered PSI matrix for the cor function
        cor_i_matrix = gene_i_matrix.corr(method="pearson")  # compute correlation which takes care of NA values
        cor_i_matrix_np = cor_i_matrix.to_numpy()  # to work with downstream steps, we need np array of cor matrix
        cor_i_matrix_np[np.triu_indices(cor_i_matrix_np.shape[0], 0)] = np.nan  # set upper triangular cor matrix to be NA because its symmetrical matrix
        idx_cor_i_binary = np.absolute(cor_i_matrix_np) > corr_threshold  # find indices/logical np array for values above threshold for cor values
        idx_cor_i_binary = 1 * idx_cor_i_binary  # binarize T/F matrix for easy column/row sum calculations
        sum_idx_cor_i_binary = np.nansum(idx_cor_i_binary, axis=0)  # determine the column sum of the binarized matrix
        # greater_than_zero_sums = sum_idx_cor_i_binary[sum_idx_cor_i_binary > 0]

        if any(sum_idx_cor_i_binary > 0):  # check if there are any correlated events,i.e., if any col sum is > 0
            if any(sum_idx_cor_i_binary == 1):  # check if there are exactly two events correlated with each other
                for col_idx in np.where(sum_idx_cor_i_binary == 1)[0]:  # from col idx we get ONE event out of the 2 correlated events
                    row_idx = np.where(idx_cor_i_binary[:, col_idx] == 1)[0][0]  # get the second event correlated with col idx event

                    num_of_nas_col_idx = gene_i_matrix.iloc[:, col_idx].isna().sum()  # determine # of NAs for the col idx event
                    num_of_nas_row_idx = gene_i_matrix.iloc[:, row_idx].isna().sum()  # determine # of NAs for the row idx event

                    number_of_nas = np.array([num_of_nas_col_idx, num_of_nas_row_idx])
                    event_to_keep = gene_i_matrix.columns.values[col_idx] if np.argmin(number_of_nas) == 0 else gene_i_matrix.columns.values[row_idx]  # keep the event with less # of NAs

                    # filtered_rows_i = event_to_keep #keep that event for this run
                    # filtered_rows = [filtered_rows, filtered_rows_i] # add that event to the initialized empty list
                    filtered_rows.append(event_to_keep)
                sum_idx_cor_i_binary = sum_idx_cor_i_binary[:len(sum_idx_cor_i_binary) - 1]  # last col is always full of NAs but the binarized matrix shows NAs to be 0 so we exclude the last col
                events_to_keep = gene_i_matrix.columns.values[np.where(sum_idx_cor_i_binary == 0)[0]]  # two events with 0 value means they are mutually exclusive so we retain that/those event(s)
                for j in events_to_keep:  # DO FOR LOOP FOR EACH EVENT IN EVENTS TO KEEP AND DO APPEND FOR THAT I
                    filtered_rows.append(j)

            else:
                cols_to_remove = np.where(np.nansum(idx_cor_i_binary, axis=0) > 0)[0]  # set of highly correlated events
                rows_to_remove = np.where(np.nansum(idx_cor_i_binary, axis=1) > 0)[0]  # prev step would not include the last col so we do the same for rows

                events_to_compare = np.union1d(cols_to_remove, rows_to_remove)  # identify the collected set of highly correlated events

                number_of_nas = gene_i_matrix.iloc[:, events_to_compare].isna().sum()  # determine the # of NAs in those events
                idx_to_keep = np.argmin(number_of_nas)  # find the event with least # of NAs
                event_to_keep = gene_i_matrix.columns.values[events_to_compare[idx_to_keep]]  # keep that event as the representative of that set of cor events
                filtered_rows.append(event_to_keep)
                cols_to_keep = np.where(np.nansum(idx_cor_i_binary, axis=0) == 0)[0]  # determine mutually exclusive events from cols
                rows_to_keep = np.where(np.nansum(idx_cor_i_binary, axis=1) == 0)[0]  # determine mutually exclusive events from rows to include for last col
                idxs_to_keep = np.intersect1d(cols_to_keep, rows_to_keep)  # here we have intersect as opposed to union because otherwise it will include repetitive events from correlated events
                # only if an event has a sum of 0 across both row and columns will it be an exclusive event
                events_to_keep = gene_i_matrix.columns.values[idxs_to_keep]
                for k in events_to_keep:  # DO FOR LOOP FOR EACH EVENT IN EVENTS TO KEEP AND DO APPEND FOR THAT I
                    filtered_rows.append(k)
        else:  # if the column sum of binarized matrix is 0 for all events then all events are mutually exclusive so we save them
            filtered_rows_i = gene_i_matrix.columns.values.tolist()
            for m in filtered_rows_i:  # DO FOR LOOP FOR EACH EVENT IN EVENTS TO KEEP AND DO APPEND FOR THAT I
                filtered_rows.append(m)
    # print("Number of filtered rows: " + str(len(filtered_rows)-1))

    return filtered_rows



# import altanalyze3.components.oncosplice.main_functions as oc
import spectra.main_functions as oc


psi_file_path = '~/psi_matrix.txt'

psi, metadata = oc.read_psi_file(psi_file_path, n_metadata_cols=11)

variable_features, psi, metadata = oc.oncosplice_feature_selection(psi, metadata,
                                                                   fold_threshold=0.2, samples_differing=4,
                                                                   corr_threshold_intercorr=0.2, corr_n_events=10,
                                                                   corr_threshold=0.8, write_files=False, savedir=None)

final_clusters, de_results = oc.oncosplice(psi=psi, metadata=metadata, variable_features=variable_features,
                                           pca_corr_threshold=0.4, npcs=30, rank='k30', force_broad='on', min_group_size=5,
                                           dPSI=0.1, dPSI_p_val=0.05, min_differential_events=100,
                                           top_n_differential_events=150, conservation='stringent', depletion_corr_threshold=0.4,
                                           speed='og', n_rounds=3, write_files=False, savedir=None)




import pandas as pd
import time
from preprocess import *

def downsample_dataframe(df, start=1000, step=1000):
    num_rows = df.shape[0]
    downsampled_dfs = {}

    for i in range(start, num_rows + 1, step):
        downsampled_dfs[i] = df.sample(n=i, random_state=42).reset_index(drop=True)

    return downsampled_dfs


# Assuming `df` is your dataframe
full_psi = pd.read_csv("/Users/tha8tf/MyProjects/Hs_RNASeq_top_alt_junctions-PSI_EventAnnotation-367-Leucegene_88kevents.txt", sep="\t", index_col=0)
downsampled_dfs = downsample_dataframe(df=full_psi, start=1000, step=1000)

# calculate time required for each downsampled dataframe using the original and the faster algorithm
time_times = {}
cpu_times = {}


for num_rows, downsampled_df in downsampled_dfs.items():
    # read the psi file in and separate out the metadata columns from numeric PSI data
    formatted_psi_file = downsampled_df.iloc[:, 7:]
    metadata = downsampled_df.iloc[:, :7]
    metadata.index = formatted_psi_file.index
    formatted_psi_file = formatted_psi_file.dropna(how='all')
    metadata = metadata.loc[formatted_psi_file.index, :]

    st = time.time()
    metadata = variance_based_feature_selection(formatted_psi_file, metadata, fold_threshold=0.3, samples_differing=4)
    et = time.time()
    el_t = (et-st) / 60
    metadata = metadata.loc[metadata['high_variance'] == True, :]
    formatted_psi_file = formatted_psi_file.loc[metadata.index, :]

    st = time.time()
    downsampled_df = fast_intercorrelation_based_feature_selection(downsampled_df, corr_threshold=0.4, corr_n_events=10)
    et = time.time()
    time_times[num_rows] = et - st

    st = time.process_time()
    downsampled_df = fast_intercorrelation_based_feature_selection(downsampled_df, corr_threshold=0.4, corr_n_events=10)
    et = time.process_time()
    cpu_times[num_rows] = et - st

# make dataframe from the time and cpu times
time_df = pd.DataFrame.from_dict(time_times, orient='index', columns=['time'])
cpu_df = pd.DataFrame.from_dict(cpu_times, orient='index', columns=['cpu_time'])

# save the dataframes to csv
time_df.to_csv("time_times_fast_corr_setting5.csv")
cpu_df.to_csv("cpu_times_fast_corr_setting5.csv")




def median_impute(df):
    """
    Impute missing values in a DataFrame with the median value of the respective row using a vectorized approach.

    Parameters:
    df (pandas.DataFrame): The input DataFrame with missing values.

    Returns:
    pandas.DataFrame: DataFrame with missing values imputed using row-wise medians.
    """

    # Define a custom function to impute missing values in each row with the row's median
    def impute_row(row):
        return row.fillna(row.median())

    # Apply the custom function to each row in the DataFrame
    imputed_df = df.apply(impute_row, axis=1)

    return imputed_df

# Example usage:
# Assuming you have a DataFrame called 'your_dataframe' with missing values
# imputed_df = medianImpute(your_dataframe)



import pandas as pd
import numpy as np
from sklearn.decomposition import PCA
from itertools import chain


def get_events_from_pca(formatted_imp_psi_file, formatted_psi_file, filtered_events, corr_threshold=0.4, n_components=30):

    # Do PCA on the list of events provided by user
    pc_psi_file = formatted_imp_psi_file.filter(items=filtered_events, axis=0)

    pca_obj = PCA(n_components, svd_solver='full')

    pca_obj.fit_transform(pc_psi_file.transpose())
    loadings = pd.DataFrame(pca_obj.components_.T, index=pc_psi_file.index)
    correlated_events_all_pc = []

    for pc in range(1, n_components + 1):
        #print(pc)
        pc_i = pd.Series(loadings.loc[:, pc - 1])
        pc_i_sorted = pc_i.sort_values(ascending=False)

        pc_top200 = pc_i_sorted.head(200).index.tolist()  # top 200 events
        pc_bottom200 = pc_i_sorted.tail(200).index.tolist()  # bottom 200 events

        pc_400events = pc_top200 + pc_bottom200

        psi_400_events_pc = formatted_psi_file.loc[pc_400events, :]

        cor_psi_400events_matrix = psi_400_events_pc.T.corr(method="pearson")
        cor_psi_400events_matrix = np.tril(cor_psi_400events_matrix, k=-1)
        idx_cor_psi_400events_binary = abs(cor_psi_400events_matrix) > corr_threshold
        idx_cor_psi_400events_binary = idx_cor_psi_400events_binary.astype(int)
        idx_cor_psi_400events_binary = pd.DataFrame(idx_cor_psi_400events_binary)
        idx_cor_psi_400events_binary.index = psi_400_events_pc.index.tolist()
        idx_cor_psi_400events_binary.columns = psi_400_events_pc.index.tolist()

        z = idx_cor_psi_400events_binary.sum(axis=0)
        z = z[z > 0].index.tolist()
        z2 = idx_cor_psi_400events_binary.sum(axis=1)
        z2 = z2[z2 > 0].index.tolist()

        correlated_events_pc = list(set(z) & set(z2))
        correlated_events_all_pc.append(correlated_events_pc)

    correlated_events_all_pc_unique = list(chain.from_iterable(correlated_events_all_pc))
    correlated_events_all_pc_unique = list(set(correlated_events_all_pc_unique))

    return correlated_events_all_pc_unique



# correvents_allpc_30pc = []
    #
    # for pc in range(1, ncomps):
    #     pci = principal_df.iloc[:, pc]
    #     pcsorted = pci.sort_values(ascending=False)
    #
    #     pc_top200 = pcsorted.head(200)
    #     pc_bottom200 = pcsorted.tail(200)
    #
    #     # pc_400events = pd.concat(pc_top200,pc_bottom200,axis=0)
    #     pc_400events = pd.DataFrame(pc_top200.append(pc_bottom200, ignore_index=False))
    #     psi_400_events_pc = formatted_psi_file.filter(items=pc_400events.index, axis=0)
    #     cor_psi_400events_matrix = psi_400_events_pc.transpose().corr(method="pearson")
    #     cor_psi_400events_matrix = cor_psi_400events_matrix.to_numpy()
    #     cor_psi_400events_matrix[np.triu_indices(cor_psi_400events_matrix.shape[0], 0)] = np.nan
    #     idx_cor_psi_400events_binary = abs(cor_psi_400events_matrix) > corr_threshold
    #     idx_cor_psi_400events_binary = 1 * idx_cor_psi_400events_binary
    #
    #     z = np.where(np.nansum(idx_cor_psi_400events_binary, axis=0) > 0)[0]
    #     z2 = np.where(np.nansum(idx_cor_psi_400events_binary, axis=1) > 0)[0]
    #
    #     correvents_pc_idx = np.intersect1d(z, z2)
    #     correvents_pc = psi_400_events_pc.index[correvents_pc_idx]
    #     correvents_allpc_30pc.append(correvents_pc)

    # correvents_allpc_30pc_unique = np.unique(correvents_allpc_30pc)




import numpy as np
import nimfa

# write the pseudocode here ##
'''
Notes:
1. the input for the nmf analysis cannot contain NA values so we use imputed PSI matrix 
2. Doing NMF on a matrix (in this case the imputed PSI matrix) decomposes the input matrix into W (basis matrix of size # of samples by rank/components) and H (coefficient matrix) matrices 
3. we will consider W matrix for all downstream analyses


Main steps in the run_nmf function: 
1. Run NMF and get the W matrix with "rank" parameter indicating the number of clusters NMF should determine. The rank parameter is determined before this analysis. 
2. Binarize W matrix in the following way so that it indicates the cluster membership of a sample (1 indicates the sample belongs to that NMF Cluster): 
    
3. Remove redundant components/clusters in the following way:
4. Finally, refine the clusters in either a "stringent" or a "conservative" way:
    a. if "stringent" 
    b. if "conservative":
'''


def run_nmf(imputed_psi_file, rank, strategy="stringent", sum_threshold=0.10):

    imputed_psi_file = imputed_psi_file.to_numpy()
    imputed_psi_file_t = imputed_psi_file.transpose()
    # can values be greater than 1 after imp? no
    # sd cannot be 0 across all rows/cols? ##what does nimfa snmf do when some columns or even rows are all 0s? ## see source for their checks
    # enter try statement here
    w = None
    try:
        nmf = nimfa.Snmf(imputed_psi_file_t, seed="nndsvd", rank=int(rank), max_iter=20, n_run=5, track_factor=True)  # n_run was originally 10 to keep the results stable
        nmf_fit = nmf()
        w = nmf_fit.basis()
        # w_a = np.array(w)  # maybe update the variable name here???
        # h = nmf_fit.coef()
        # H = np.array(h)
    except ValueError: 
        w = imputed_psi_file # this needs to change -- how are you taking care of errors?

    finally:
        nmf_matrix = w

    if int(rank) == 2:
        binarized_output = binarize_nmf(w, sum_threshold, par=1, lower_bound=3, upper_bound=40)  # adjusted way of determining outliers with par = 1
    else:
        binarized_output = binarize_nmf(w, sum_threshold, par=2, lower_bound=3, upper_bound=40)  # standard way of determining the outliers

    binarized_output_for_refining = np.array(binarized_output)
    
    # what if binarized output is all 0's? can reduce the the sum_threshold in the except statement within the main function call or if it is user provided, print the info and let the user know about it
    refined_binarized_output = remove_redundant_components(binarized_output_for_refining)

    # refined_binarized_output = finalize_binarized_output(refined_binarized_output, strategy=strategy) # this line not needed here -- MV does nothing in this function in original code
    # ...contd. from MV's dissertation, I believe this is a part of the ExpandSampleClusters/LinearSVC stuff so ignore it here for sure.

    return nmf_matrix, binarized_output, refined_binarized_output


def binarize_nmf(w, sum_threshold, par=1, lower_bound=3, upper_bound=40):

    w = w.transpose()
    # components = w.shape[0]  # this is number of rows/components
    samples = w.shape[1]  # number of columns/samples
    
    ## if there are more than 40 samples that have a value of 0.1 in the W matrix OR if very very few samples (less than 3) have a value of 0.1, 
    ## then the allocation to a cluster is determined by doing mean+par*std for a cluster
    z_sum_threshold = np.where(w > sum_threshold, 1, 0)  # binarize W matrix based on sum_threshold (greater than sum_threshold (0.1) means 1)
    num_vec = np.sum(z_sum_threshold, axis=1)  # for every component/row, number of samples that have a value greater than sum_threshold (0.1)

    compstd_vec_criteria1 = num_vec < lower_bound 
    compstd_vec_criteria2 = num_vec > upper_bound
    compstd_vec = np.where(np.logical_or(compstd_vec_criteria1, compstd_vec_criteria2) == True) # 40 can be a fraction of the num value #compstd is True if value in num_vec is < 3 or > 40

    # compstd_vec = np.where((num_vec < 3 or num_vec > 40), True, False)
    subset_compstd_w = w[compstd_vec, :].reshape(np.shape(compstd_vec)[1], samples)  # reshape it into a matrix of # of components (where compstd_vec == TRUE) by # of samples
    mean_vec = np.mean(subset_compstd_w, axis=1).transpose()
    st_vec = np.std(subset_compstd_w, axis=1).transpose()

    for i in range(samples):
        #print(i)
        z_sum_threshold[compstd_vec, i] = np.where((w[compstd_vec, i] >= mean_vec + par*st_vec), 1, 0)

    binarized_output = z_sum_threshold

    return binarized_output


def remove_redundant_components(binarized_output):

    components = binarized_output.shape[0]  # this is number of rows/components
    refined_binarized_output = []

    for i in range(components):
        component1 = binarized_output[i, :]
        sum1 = sum(component1) # number of samples belonging to the component/cluster
        flag = False
        component1_membership = np.where(component1 == 1) # samples that belong to the i component
        for j in range(components):
            if i != j:
                component2 = binarized_output[j, :]
                
                sum_of_overlapping_labels = sum(component2[component1_membership])
                sum2 = sum(component2)
                # print(float(sum_of_overlapping_labels) / sum1)
                if float(sum_of_overlapping_labels)/sum1 > 0.5:
                    if sum2 > sum1:
                        flag = True

        if not flag:
            refined_binarized_output.append(component1)

    refined_binarized_output = np.array(refined_binarized_output)

    return refined_binarized_output


def finalize_binarized_output(refined_binarized_output, strategy="stringent"):

    if strategy == "stringent":
        print("stringent")
    else:
        print("conservative")

    return refined_binarized_output
